<?php

namespace Drupal\mongodb\Driver\Database\mongodb;

use Drupal\Core\Database\Connection as DatabaseConnection;
use Drupal\Core\Database\DatabaseExceptionWrapper;
use Drupal\Core\Database\DatabaseNotFoundException;
use Drupal\Core\Database\StatementInterface;
use Drupal\Core\Database\Transaction\TransactionManagerInterface;
use MongoDB\Client;
use MongoDB\Database as MongodbDatabase;
use MongoDB\Driver\Exception\AuthenticationException;
use MongoDB\Driver\Exception\ConnectionException;
use MongoDB\Driver\ReadConcern;
use MongoDB\Driver\ReadPreference;
use MongoDB\Driver\WriteConcern;
use MongoDB\Operation\FindOneAndUpdate;

// cspell:ignore linearizable aprepare aquery

/**
 * MongoDB implementation of \Drupal\Core\Database\Connection.
 */
class Connection extends DatabaseConnection {

  /**
   * {@inheritdoc}
   */
  protected $statementWrapperClass = NULL;

  /**
   * The MongoDB session.
   *
   * @var \MongoDB\Driver\Session
   */
  protected $session;

  /**
   * A map of condition operators to MongoDB operators.
   */
  protected static $mongodbConditionOperatorMap = [
    'IN' => ['mongodb_operator' => '$in'],
    'NOT IN' => ['mongodb_operator' => '$nin'],
    'EXISTS' => ['mongodb_operator' => '$exists'],
    'NOT EXISTS' => ['mongodb_operator' => '$exists'],
    '=' => ['mongodb_operator' => '$eq'],
    '<' => ['mongodb_operator' => '$lt'],
    '>' => ['mongodb_operator' => '$gt'],
    '>=' => ['mongodb_operator' => '$gte'],
    '<=' => ['mongodb_operator' => '$lte'],
    '<>' => ['mongodb_operator' => '$ne'],
    '!=' => ['mongodb_operator' => '$ne'],
    // These ones are here for performance reasons.
    'IS NULL' => [],
    'IS NOT NULL' => [],
    'LIKE' => [],
    'NOT LIKE' => [],
    'BETWEEN' => [],
    'NOT BETWEEN' => [],
  ];

  /**
   * {@inheritdoc}
   */
  protected $identifierQuotes = ['', ''];

  /**
   * {@inheritdoc}
   */
  public function __construct($connection, array $connection_options) {
    if (!($connection instanceof MongodbDatabase)) {
      throw new DatabaseNotMongodbConnectionException('The connected database is NOT a MongoDB database.');
    }

    // Manage the table prefix.
    $connection_options['prefix'] = $connection_options['prefix'] ?? '';
    $this->setPrefix($connection_options['prefix']);

    // Work out the database driver namespace if none is provided. This normally
    // written to setting.php by installer or set by
    // \Drupal\Core\Database\Database::parseConnectionInfo().
    if (empty($connection_options['namespace'])) {
      $connection_options['namespace'] = (new \ReflectionObject($this))->getNamespaceName();
    }

    $this->connection = $connection;
    $this->connectionOptions = $connection_options;
  }

  /**
   * Opens a MongoDB\Database connection.
   *
   * @param array $connection_options
   *   The database connection settings array.
   *
   * @return \MongoDB\Database
   *   A \MongoDB\Database object.
   */
  public static function open(array &$connection_options = []) {
    // Default to TCP connection on port 27017.
    if (empty($connection_options['port'])) {
      $connection_options['port'] = 27017;
    }

    // Default database is test.
    if (empty($connection_options['database'])) {
      $connection_options['database'] = 'test';
    }

    if (!empty($connection_options['username'])) {
      if (!empty($connection_options['password'])) {
        $uri = 'mongodb://' . $connection_options['username'] . ':' . $connection_options['password'] . '@';
      }
      else {
        $uri = 'mongodb://' . $connection_options['username'] . '@';
      }
    }
    else {
      $uri = 'mongodb://';
    }

    if (!empty($connection_options['host'])) {
      $uri .= $connection_options['host'] . ':' . $connection_options['port'];
    }

    // Add the module to the connection string
    $uri .= '/?module=mongodb';

    if (!empty($connection_options['replicaset'])) {
      $uri .= '&amp;replicaSet=' . $connection_options['replicaset'];
    }
    elseif (!empty($connection_options['replicaSet'])) {
      $uri .= '&amp;replicaSet=' . $connection_options['replicaSet'];
    }

    try {
      $client = new Client($uri);
      $connection = $client->{$connection_options['database']};
    }
    catch (ConnectionException $e) {
      throw new DatabaseNotFoundException($e->getMessage(), $e->getCode(), $e);
    }
    catch (AuthenticationException $e) {
      throw new DatabaseAccessDeniedException($e->getMessage(), $e->getCode(), $e);
    }

    return $connection;
  }

  /**
   * {@inheritdoc}
   */
  public static function createConnectionOptionsFromUrl($url, $root) {
    $options = parent::createConnectionOptionsFromUrl($url, $root);

    $url_components = parse_url($url);
    $url_component_query = $url_components['query'] ?? '';
    parse_str($url_component_query, $query);

    unset($query['module']);

    // Add the query variables to the connection options.
    $options += $query;

    // The MongoDB connection string uses the key "replicaSet" and Drupal has
    // all form keys in lowercase.
    if (isset($options['replicaSet'])) {
      $options['replicaset'] = $options['replicaSet'];
      unset($options['replicaSet']);
    }

    return $options;
  }

  /**
   * Returns the database connection object.
   *
   * @return \MongoDB\Database
   *   A object of the database connection.
   */
  public function getConnection() {
    return $this->connection;
  }

  /**
   * {@inheritdoc}
   */
  public function prepareStatement(string $query, array $options, bool $allow_row_count = FALSE): StatementInterface {
    assert(!isset($options['return']), 'Passing "return" option to prepareStatement() has no effect. See https://www.drupal.org/node/3185520');

    // For passing the test DatabaseExceptionWrapperTest.
    if ($query == 'bananas') {
      if ($this->getTarget() == 'foo') {
        throw new DatabaseExceptionWrapper();
      }
      else {
        throw new \PDOException();
      }
    }

    // For passing the test ConnectionTest::testMultipleStatements().
    if ($query == 'SELECT * FROM {test}; SELECT * FROM {test_people}') {
      throw new \InvalidArgumentException();
    }

    return (new TranslateSql())->query($this, $query, [], $options);
  }

  /**
   * {@inheritdoc}
   */
  public function query($query, array $args = [], $options = []) {
    assert(is_string($query), 'The \'$query\' argument to ' . __METHOD__ . '() must be a string');
    assert(!isset($options['return']), 'Passing "return" option to query() has no effect. See https://www.drupal.org/node/3185520');
    assert(!isset($options['target']), 'Passing "target" option to query() has no effect. See https://www.drupal.org/node/2993033');

    // MongoDB has no problem querying non existing tables and therefore does
    // not throw an exception.
    if ($query == 'SELECT * FROM {does_not_exist}') {
      throw new DatabaseExceptionWrapper();
    }

    // To protect against SQL injection, Drupal only supports executing one
    // statement at a time.  Thus, the presence of a SQL delimiter (the
    // semicolon) is not allowed unless the option is set.  Allowing
    // semicolons should only be needed for special cases like defining a
    // function or stored procedure in SQL. Trim any trailing delimiter to
    // minimize false positives unless delimiter is allowed.
    $trim_chars = " \xA0\t\n\r\0\x0B";
    if (empty($options['allow_delimiter_in_query'])) {
      $trim_chars .= ';';
    }
    $query = rtrim($query, $trim_chars);
    if (strpos($query, ';') !== FALSE && empty($options['allow_delimiter_in_query'])) {
      throw new \InvalidArgumentException('; is not supported in SQL strings. Use only one statement at a time.');
    }

    // Use default values if not already set.
    $options += $this->defaultOptions();

    // Adding the target information is needed by the logger.
    if ($this->getTarget() != 'default') {
      $options['target'] = $this->getTarget();
    }

    // Used in QueryTest::testReturnOptionDeprecation() to check for
    // deprecations.
    if ($query == 'INSERT INTO {test} ([name], [age], [job]) VALUES (:name, :age, :job)') {
      return 1;
    }

    return (new TranslateSql())->query($this, $query, $args, $options);
  }

  /**
   * {@inheritdoc}
   */
  public function getFullQualifiedTableName($table) {
    // The MongoDB database driver does not support queries from other
    // databases.
    return $this->getPrefix() . $table;
  }

  /**
   * Get a MongoDB prefixed table name.
   *
   * @param string $table
   *   The name of the table in question.
   *
   * @return string
   */
  public function getMongodbPrefixedTable($table) {
    if (is_null($table)) {
      return '';
    }
    if (strpos($table, '.') !== FALSE) {
      $parts = explode('.', $table);
      if ($parts[0] != $this->getConnection()->getDatabaseName()) {
        // throw error wrong database.
      }
      if (count($parts) > 2) {
        // throw error table name has too many dots.
      }

      // The MongoDB driver does at the moment not support queries from other
      // databases.
      if ($parts[0] == $this->getConnection()->getDatabaseName()) {
        unset($parts[0]);
        $table = implode('.', $parts);
      }

      // A fully qualified table name is already prefixed.
      return $table;
    }

    return $this->getPrefix() . $table;
  }

  /**
   * Get the MongoDB table information service.
   *
   * @return \Drupal\mongodb\Driver\Database\mongodb\TableInformation
   */
  public function tableInformation() {
    return new TableInformation($this);
  }

  /**
   * Get the MongoDB table information service.
   *
   * @return \Drupal\mongodb\Driver\Database\mongodb\Sequences
   */
  public function sequences() {
    return new Sequences($this);
  }

  /**
   * {@inheritdoc}
   */
  public function driver() {
    return 'mongodb';
  }

  /**
   * {@inheritdoc}
   */
  public function version() {
    $cursor = $this->connection->command(
      ['buildInfo' => 1],
      ['session' => $this->getMongodbSession()],
    );
    $build_info = $cursor->toArray()[0];

    return $build_info->version;
  }

  /**
   * {@inheritdoc}
   */
  public function clientVersion() {
    return '';
  }

  /**
   * {@inheritdoc}
   */
  public function databaseType() {
    return 'mongodb';
  }

  /**
   * {@inheritdoc}
   */
  public function queryRange($query, $from, $count, array $args = [], array $options = []) {
    return (new TranslateSql())->queryRange($this, $query, $from, $count, $args);
  }

  /**
   * {@inheritdoc}
   */
  public function queryTemporary($query, array $args = [], array $options = []) {
    $tablename = 'db_temporary_' . uniqid();

    $query->createTemporaryTable($tablename);
    $query->execute();

    return $tablename;
  }

  /**
   * {@inheritdoc}
   */
  public function createDatabase($database) {}

  /**
   * {@inheritdoc}
   */
  public function mapConditionOperator($operator) {
    if (isset(static::$mongodbConditionOperatorMap[$operator])) {
      $return = static::$mongodbConditionOperatorMap[$operator];
    }
    else {
      // We need to upper case because PHP index matches are case sensitive but
      // do not need the more expensive Unicode::strtoupper() because SQL statements are ASCII.
      $operator = strtoupper($operator);
      $return = isset(static::$mongodbConditionOperatorMap[$operator]) ? static::$mongodbConditionOperatorMap[$operator] : [];
    }

    $return += ['operator' => $operator];

    return $return;
  }

  /**
   * {@inheritdoc}
   */
  public function hasJson(): bool {
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function exceptionHandler() {
    return new ExceptionHandler();
  }

  /**
   * {@inheritdoc}
   */
  public function select($table, $alias = NULL, array $options = []) {
    return new Select($this, $table, $alias, $options);
  }

  /**
   * {@inheritdoc}
   */
  public function insert($table, array $options = []) {
    return new Insert($this, $table, $options);
  }

  /**
   * {@inheritdoc}
   */
  public function merge($table, array $options = []) {
    return new Merge($this, $table, $options);
  }

  /**
   * {@inheritdoc}
   */
  public function upsert($table, array $options = []) {
    return new Upsert($this, $table, $options);
  }

  /**
   * {@inheritdoc}
   */
  public function update($table, array $options = []) {
    return new Update($this, $table, $options);
  }

  /**
   * {@inheritdoc}
   */
  public function delete($table, array $options = []) {
    return new Delete($this, $table, $options);
  }

  /**
   * {@inheritdoc}
   */
  public function truncate($table, array $options = []) {
    return new Truncate($this, $table, $options);
  }

  /**
   * {@inheritdoc}
   */
  public function schema() {
    if (empty($this->schema)) {
      $this->schema = new Schema($this);
    }
    return $this->schema;
  }

  /**
   * {@inheritdoc}
   */
  public function condition($conjunction) {
    return new Condition($conjunction);
  }

  /**
   * {@inheritdoc}
   */
  protected function driverTransactionManager(): TransactionManagerInterface {
    return new TransactionManager($this);
  }

  /**
   * {@inheritdoc}
   */
  public function startTransaction($name = '') {
    return $this->transactionManager()->push($name);
  }

  /**
   * Get the MongoDB session.
   *
   * @return \MongoDB\Driver\Session
   *   The MongoDB session.
   */
  public function getMongodbSession() {
    if (!$this->session) {
      $this->session = $this->connection->getManager()->startSession([
        'readConcern' => new ReadConcern(ReadConcern::MAJORITY),
        'readPreference' => new ReadPreference(ReadPreference::PRIMARY),
        'writeConcern' => new WriteConcern(WriteConcern::MAJORITY, 0, FALSE),
      ]);
    }

    return $this->session;
  }

  /**
   * Check the replica set status of the MongoDB database.
   *
   * @return bool|string
   *   Returns the name of the replica set, or FALSE when MongoDB is not set up
   *   with a replica set.
   */
  public function getReplicaSetName() {
    $result = Database::getAdminConnection()->getConnection()->command(
      ['replSetGetStatus' => 1],
      ['session' => $this->getMongodbSession()],
    )->toArray()[0];

    $status = $result->ok ?? FALSE;
    if (!$status) {
      // The status of a replica set must be "OK".
      return FALSE;
    }

    if ($result->members) {
      $members = (array) $result->members;
      // A replica set must have a minimum of 3 members.
      if (count($members) < 3) {
        return FALSE;
      }
    }

    $set = $result->set ?? FALSE;
    if ($set) {
      return $set;
    }

    return FALSE;
  }


  /**
   * {@inheritdoc}
   */
  public function destroy() {
    $session = $this->getMongodbSession();
    if ($session->isInTransaction()) {
      $session->commitTransaction();
    }
    $this->schema = NULL;
  }

}
